/* 
 * File:   NetStructures.h
 * Author: Peter G. Jensen
 *
 * Created on 09 March 2016, 21:08
 */

#ifndef NETSTRUCTURES_H
#define NETSTRUCTURES_H

#include <limits>
#include <vector>
#include <iostream>

namespace PetriEngine {

    struct Arc {
        uint32_t place;
        uint32_t weight;
        bool skip = false;
        bool inhib = false;
        bool transport = false;
        uint32_t lowerBound;
        uint32_t upperBound;
        bool lowerStrict;
        bool upperStrict;
        std::string interval;


        Arc() :
        place(std::numeric_limits<uint32_t>::max()),
        weight(std::numeric_limits<uint32_t>::max()),
        skip(false),
        inhib(false),
        lowerBound(std::numeric_limits<uint32_t>::max()),
        upperBound(std::numeric_limits<uint32_t> ::max()),
        lowerStrict(false),
        upperStrict(false)
        {
        };
        
        bool operator < (const Arc& other) const
        {
            return place < other.place;
        }
        
        bool operator == (const Arc& other) const
        {
            return place == other.place && weight == other.weight && inhib == other.inhib;
        }
    };

    struct Transition {
        std::vector<Arc> pre;
        std::vector<Arc> post;
        bool skip = false;
        bool inhib = false;
        bool urgent = false;
    };

    struct Place {
        std::vector<uint32_t> consumers; // things consuming
        std::vector<uint32_t> producers; // things producing
       // Colored::TimeInvariant timeinvariant;
        bool skip = false;
        bool inhib = false;
    };
}
#endif /* NETSTRUCTURES_H */

